package envoy.client.data.commands;

import java.util.*;
import java.util.function.Function;
import java.util.logging.Level;
import java.util.stream.Collectors;

import envoy.util.EnvoyLog;

/**
 * This class stores all {@link SystemCommand}s used.
 * <p>
 * Project: <strong>envoy-client</strong><br>
 * File: <strong>SystemCommandsMap.java</strong><br>
 * Created: <strong>17.07.2020</strong><br>
 *
 * @author Leon Hofmeister
 * @since Envoy Client v0.1-beta
 * @apiNote Please refrain from using the "/"-char in keys of this map.
 *          Unexpected behavior will occur.
 */
public class SystemCommandsMap {

	private final HashMap<String, SystemCommand> systemCommands = new HashMap<>();

	/**
	 * @param command           the string that must be inputted to execute the
	 *                          given action
	 * @param action            the action that should be performed
	 * @param numberOfArguments the amount of arguments that need to be parsed for
	 *                          the underlying function
	 * @since Envoy Client v0.1-beta
	 */
	public void addCommand(String command, Function<String[], Void> action, int numberOfArguments) {
		systemCommands.put(command, new SystemCommand(action, numberOfArguments, ""));
	}

	/**
	 * @param command           the string that must be inputted to execute the
	 *                          given action
	 * @param action            the action that should be performed
	 * @param numberOfArguments the amount of arguments that need to be parsed for
	 *                          the underlying function
	 * @param description       the description of this {@link SystemCommand}
	 * @since Envoy Client v0.1-beta
	 */
	public void addCommand(String command, Function<String[], Void> action, int numberOfArguments, String description) {
		systemCommands.put(command, new SystemCommand(action, numberOfArguments, description));
	}

	/**
	 * Adds a new {@link SystemCommand} to the map that does not depend upon
	 * arguments in the given String
	 *
	 * @param command the string that must be inputted to execute the given action
	 * @param action  the action that should be performed. To see why this Function
	 *                takes a {@code String[]}, see {@link SystemCommand}
	 * @since Envoy Client v0.1-beta
	 */
	public void addNoArgCommand(String command, Function<String[], Void> action) { addCommand(command, action, 0); }

	/**
	 * Adds a new {@link SystemCommand} to the map that does not depend upon
	 * arguments in the given String
	 *
	 * @param command     the string that must be inputted to execute the given
	 *                    action
	 * @param action      the action that should be performed. To see why this
	 *                    Function takes a {@code String[]}, see
	 *                    {@link SystemCommand}
	 * @param description the description of this {@link SystemCommand}
	 * @since Envoy Client v0.1-beta
	 */
	public void addNoArgCommand(String command, Function<String[], Void> action, String description) { addCommand(command, action, 0); }

	/**
	 * @param command           the string that must be inputted to execute the
	 *                          given action
	 * @param action            the action that should be performed. To see why this
	 *                          Function takes a {@code String[]}, see
	 *                          {@link SystemCommand}
	 * @param numberOfArguments the amount of arguments that need to be parsed for
	 *                          the underlying function
	 * @since Envoy Client v0.1-beta
	 */
	public void add(String command, Function<String[], Void> action, int numberOfArguments) {
		systemCommands.put(command, new SystemCommand(action, numberOfArguments, ""));
	}

	/**
	 * This method checks if the input String is a key in the map and returns the
	 * wrapped System command if present.
	 * Its intended usage is after a "/" has been detected in the input String.
	 * It will return an empty optional if the value after the slash is not a key in
	 * the map, which is a valid case (i.e. input="3/4" and "4" is not a key in the
	 * map).
	 * <p>
	 * Usage example:<br>
	 * {@code SystemCommandsMap systemCommands = new SystemCommandsMap();}<br>
	 * {@code systemCommands.add("example", Function.identity, 1);}<br>
	 * {@code ....}<br>
	 * user input: {@code "/example xyz ..."}<br>
	 * {@code systemCommands.checkPresent("example xyz ...")}
	 * result: {@code SystemCommand[action=Function.identity, numberOfArguments=1]}
	 *
	 * @param input the input string given by the user, <b>excluding the "/"</b>
	 * @return the wrapped system command, if present
	 * @since Envoy Client v0.1-beta
	 */
	public Optional<SystemCommand> checkPresent(String input) {
		final var firstWord = input.substring(0, input.indexOf(" "));
		return Optional.ofNullable(systemCommands.get(firstWord));
	}

	/**
	 * This method checks if the input String is a key in the map and executes the
	 * wrapped System command if present.
	 * Its intended usage is after a "/" has been detected in the input String.
	 * It will do nothing if the value after the slash is not a key in
	 * the map, which is a valid case (i.e. input="3/4" and "4" is not a key in the
	 * map).
	 * <p>
	 * Usage example:<br>
	 * {@code SystemCommandsMap systemCommands = new SystemCommandsMap();}<br>
	 * {@code systemCommands.add("example", (words)-> &lt;Button&gt;.setText(words[0]), 1);}<br>
	 * {@code ....}<br>
	 * user input: {@code "/example xyz ..."}<br>
	 * {@code systemCommands.executeIfPresent("example xyz ...")}
	 * result: {@code <Button>.getText()=="xyz"}
	 *
	 * @param input the input string given by the user, <b>excluding the "/"</b>
	 * @since Envoy Client v0.1-beta
	 */
	public void executeIfPresent(String input) {
		checkPresent(input).ifPresent(systemCommand -> {
			// Splitting the String so that the leading command including the first " " is
			// removed and only as many following words as allowed by the system command
			// persist
			final var remainingString = input.substring(input.indexOf(" ") + 1);
			// TODO: Current implementation will fail in certain cases, i.e. two characters
			// behind each other (" "), not enough words, ...
			final var arguments = Arrays.copyOfRange(remainingString.split(" "), 0, systemCommand.getNumberOfArguments());
			// Executing the function
			try {
				systemCommand.getAction().apply(arguments);
			} catch (final Exception e) {
				EnvoyLog.getLogger(SystemCommandsMap.class).log(Level.WARNING, "The system command " +
				// detected command
				input.substring(0, input.indexOf(" ")) + " threw an exception: ", e);
			}
		});
	}

	/**
	 * Retrieves the recommendations based on the current command entered (i.e the
	 * String "/exam" should be entered as "exam"). If none are present, nothing
	 * will be done.
	 * Otherwise the given function will be executed on the recommendations.
	 *
	 * @param input  the String between "/" and " "
	 * @param action the action that should be taken for the recommendations, if any
	 *               are present
	 * @since Envoy Client v0.1-beta
	 */
	public void requestRecommendations(String input, Function<Set<String>, Void> action) {
		final var partialCommand = input.substring(0, input.indexOf(" "));
		// Get the expected commands
		final var recommendations = recommendCommands(partialCommand);
		if (recommendations.isEmpty()) return;
		// Execute the given action
		else action.apply(recommendations);
	}

	/**
	 * Recommends commands based upon the currently entered input.<br>
	 * In the current implementation, all we check is whether a key contains this
	 * input. This might be updated later on.
	 *
	 * @param partialCommand the partially entered command
	 * @return a set of all commands that match this input
	 */
	private Set<String> recommendCommands(String partialCommand) {
		// current implementation only looks if input is contained within a command,
		// might be updated
		return systemCommands.keySet().stream().filter(command -> command.contains(partialCommand)).collect(Collectors.toSet());
	}

	/**
	 * @return all {@link SystemCommand}s used with the underlying command as key
	 * @since Envoy Client v0.1-beta
	 */
	public HashMap<String, SystemCommand> getSystemCommands() { return systemCommands; }
}
